@@ -18,11 +18,17 @@ from .message_mapping import (
 from pyvast import VAST
 import random
 from shlex import split as lexical_split
+import socket
 from string import ascii_lowercase as letters
+import sys
+from .metrics import Gauge, Summary
+from threatbus.logger import setup as setup_logging_threatbus
 from threatbus.data import Intel, IntelDecoder, Sighting, SightingEncoder, Operation
+import time
 import zmq
 
-logger = logging.getLogger(__name__)
+logger_name = "pyvast-threatbus"
+logger = logging.getLogger(logger_name)
 matcher_name = None
 # list of all running async tasks of the bridge
 async_tasks = []
@@ -30,8 +36,21 @@ async_tasks = []
 p2p_topic = None
 max_open_tasks = None
 
+# metric definitions
+metrics = []
+g_iocs_added = Gauge("added_iocs")
+g_iocs_removed = Gauge("removed_iocs")
+metrics += [g_iocs_added, g_iocs_removed]
+s_retro_matches_per_ioc = Summary("retro_matches_per_ioc")
+s_retro_query_time_s_per_ioc = Summary("retro_query_time_per_ioc")
+g_live_matcher_sightings = Gauge("live_matcher_sightings")
 
-def setup_logging(level):
+
+def setup_logging_with_level(level: str):
+    """
+    Sets up a the global logger for console logging with the given loglevel.
+    @param level The loglevel to use, e.g., "DEBUG"
+    """
     global logger
     log_level = logging.getLevelName(level.upper())
 
@@ -46,13 +65,21 @@ def setup_logging(level):
     logger.addHandler(handler)
 
 
+def setup_logging_with_config(config: confuse.Subview):
+    """
+    Sets up the global logger as configured in the `config` object.
+    @param config The user-defined logging configuration
+    """
+    global logger
+    logger = setup_logging_threatbus(config, logger_name)
+
+
 def validate_config(config: confuse.Subview):
     assert config, "config must not be None"
     config["vast"].get(str)
     config["vast_binary"].get(str)
     config["threatbus"].get(str)
     config["snapshot"].get(int)
-    config["loglevel"].get(str)
     config["retro_match"].get(bool)
     config["retro_match_max_events"].get(int)
     config["unflatten"].get(bool)
@@ -62,6 +89,17 @@ def validate_config(config: confuse.Subview):
     config["transform_context"].add(None)
     config["sink"].add(None)
 
+    # logging
+    if config["logging"]["console"].get(bool):
+        config["logging"]["console_verbosity"].get(str)
+    if config["logging"]["file"].get(bool):
+        config["logging"]["file_verbosity"].get(str)
+        config["logging"]["filename"].get(str)
+
+    # metrics
+    config["metrics"]["interval"].get(int)
+    config["metrics"]["filename"].get(str)
+
 
 def cancel_async_tasks():
     """
@@ -83,6 +121,8 @@ async def start(
     retro_match_max_events: int,
     unflatten: bool,
     max_open_files: int,
+    metrics_interval: int,
+    metrics_filename: str,
     transform_cmd: str = None,
     sink: str = None,
 ):
@@ -98,11 +138,13 @@ async def start(
     @param retro_match_max_events Max amount of retro match results
     @param unflatten Boolean flag to unflatten JSON when received from VAST
     @param max_open_files The maximum number of concurrent background tasks for VAST queries.
+    @param merics_interval The interval in seconds to bucketize metrics
+    @param metrics_filename The filename (system path) to store metrics at
     @param transform_cmd The command to use to transform Sighting context with
     @param sink Forward sighting context to this sink (subprocess) instead of
         reporting back to Threat Bus
     """
-    global logger, async_tasks, p2p_topic, max_open_tasks
+    global logger, async_tasks, p2p_topic, max_open_tasks, metrics
     # needs to be created inside the same eventloop where it is used
     max_open_tasks = asyncio.Semaphore(max_open_files)
     vast = VAST(binary=vast_binary, endpoint=vast_endpoint, logger=logger)
@@ -159,17 +201,56 @@ async def start(
         )
     )
 
-    if not retro_match:
+    if retro_match:
+        # add metrics for retro-matching to the metric output
+        metrics += [s_retro_matches_per_ioc, s_retro_query_time_s_per_ioc]
+    else:
+        # add metrics for live-matching to the metric output
+        metrics.append(g_live_matcher_sightings)
         async_tasks.append(
             asyncio.create_task(
                 live_match_vast(vast_binary, vast_endpoint, sightings_queue)
             )
         )
 
+    if metrics_interval:
+        async_tasks.append(
+            asyncio.create_task(write_metrics(metrics_interval, metrics_filename))
+        )
+
     atexit.register(cancel_async_tasks)
     return await asyncio.gather(*async_tasks)
 
 
+async def write_metrics(every: int, to: str):
+    """
+    Periodically writes metrics to a file.
+    @param every The interval to write metrics, in seconds
+    @param to the filepath to write to
+    """
+    while True:
+        line = f"pyvast-threatbus,host={socket.gethostname()}"
+        start_length = len(line)
+        for m in metrics:
+            if not m.is_set:
+                continue
+            if type(m) is Gauge:
+                line += f" {m.name}={m.value}"
+            if type(m) is Summary:
+                line += (
+                    f" {m.name}_min={m.min},{m.name}_max={m.max},{m.name}_avg={m.avg}"
+                )
+            m.reset()
+
+        if len(line) > start_length:
+            # only update the file if there were metrics collected.
+            line += f" {time.time_ns()}"  # append current nanoseconds ts
+            with open(to, "a") as f:
+                f.write(line + "\n")
+
+        await asyncio.sleep(every)
+
+
 async def receive(pub_endpoint: str, topic: str, intel_queue: asyncio.Queue):
     """
     Starts a zmq subscriber on the given endpoint and listens for new messages
@@ -224,6 +305,7 @@ async def retro_match_vast(
     @param sightings_queue The queue to put new sightings into
     @param unflatten Boolean flag to unflatten JSON when received from VAST
     """
+    start = time.time()
     query = to_vast_query(intel)
     if not query:
         return
@@ -242,6 +324,8 @@ async def retro_match_vast(
                 reported += 1
                 await sightings_queue.put(sighting)
         logger.debug(f"Retro-matched {reported} sighting(s) for intel: {intel}")
+        s_retro_matches_per_ioc.observe(reported)
+        s_retro_query_time_s_per_ioc.observe(time.time() - start)
 
 
 async def ingest_vast_ioc(vast_binary, vast_endpoint, intel):
@@ -316,6 +400,7 @@ async def match_intel(
             )
             continue
         if intel.operation == Operation.ADD:
+            g_iocs_added.inc()
             if retro_match:
                 asyncio.create_task(
                     retro_match_vast(
@@ -330,6 +415,7 @@ async def match_intel(
             else:
                 asyncio.create_task(ingest_vast_ioc(vast_binary, vast_endpoint, intel))
         elif intel.operation == Operation.REMOVE:
+            g_iocs_removed.inc()
             if retro_match:
                 continue
             asyncio.create_task(remove_vast_ioc(vast_binary, vast_endpoint, intel))
@@ -365,6 +451,7 @@ async def live_match_vast(
         if not sighting:
             logger.error(f"Cannot parse sighting-output from VAST: {vast_sighting}")
             continue
+        g_live_matcher_sightings.inc()
         await sightings_queue.put(sighting)
 
 
@@ -555,86 +642,16 @@ async def heartbeat(endpoint: str, p2p_topic: str, interval: int = 5):
 def main():
     parser = argparse.ArgumentParser()
     parser.add_argument("--config", "-c", help="Path to a configuration file")
-    parser.add_argument(
-        "--vast",
-        "-v",
-        dest="vast",
-        default="localhost:42000",
-        help="Endpoint of a running VAST node",
-    )
-    parser.add_argument(
-        "--vast-binary",
-        "-b",
-        dest="vast_binary",
-        default="vast",
-        help="The vast command to use (either absolte path or a command available in $PATH)",
-    )
-    parser.add_argument(
-        "--threatbus",
-        "-t",
-        dest="threatbus",
-        default="localhost:13370",
-        help="Management endpoint of a VAST Threat Bus node",
-    )
-    parser.add_argument(
-        "--snapshot",
-        "-s",
-        dest="snapshot",
-        default=0,
-        type=int,
-        help="Request intelligence snapshot for past n days",
-    )
-    parser.add_argument(
-        "--loglevel",
-        "-l",
-        dest="loglevel",
-        default="info",
-        help="Loglevel to use",
-    )
-    parser.add_argument(
-        "--retro-match",
-        dest="retro_match",
-        action="store_true",
-        help="Use plain vast queries instead of live-matcher",
-    )
-    parser.add_argument(
-        "--unflatten",
-        dest="unflatten",
-        action="store_true",
-        help="Only applicable when --retro-match is used. Unflatten the JSON results from VAST. The unflattening is applied immediately after retrieving the results from VAST, i.e., unflatten is applied before all further processing steps like --transform-context, --sink, or reporting back to Threat Bus.",
-    )
-    parser.add_argument(
-        "--retro-match-max-events",
-        dest="retro_match_max_events",
-        default=0,
-        type=int,
-        help="Use this options to fine-tune '--retro-match'. The app passes this option to VAST to at most return the configured number of sightings per IoC.",
-    )
-    parser.add_argument(
-        "--transform-context",
-        "-T",
-        dest="transform_context",
-        default=None,
-        help="Forward the context of each sighting (only the contents without the Threat Bus specific sighting structure) via a UNIX pipe. This option takes a command line string to use and invokes it as direct subprocess without shell / globbing support. Note: Treated as template string. Occurrences of '%%ioc' get replaced with the matched IoC.",
-    )
-    parser.add_argument(
-        "--sink",
-        "-S",
-        dest="sink",
-        default=None,
-        help="If sink is specified, sightings are not reported back to Threat Bus. Instead, the context of a sighting (only the contents without the Threat Bus specific sighting structure) is forwarded to the specified sink via a UNIX pipe. This option takes a command line string to use and invokes it as direct subprocess without shell / globbing support.",
-    )
-    parser.add_argument(
-        "--max-background-tasks",
-        "-U",
-        dest="max-background-tasks",
-        default=100,
-        type=int,
-        help="Controls the maximum number of concurrent background tasks for VAST queries. Default is 100.",
-    )
     args = parser.parse_args()
 
-    config = confuse.Configuration("pyvast-threatbus")
+    # we need to use an underscore in the configuration name "pyvast_threatbus"
+    # confuse uses the configuration name to lookup environment variables, but
+    # it simply upper-cases that name. Dashes are not replaced properly. Using a
+    # dash in the configuration name makes it impossible to configure the
+    # APPNAMEDIR env variable to overwrite search paths, i.e., in systemd
+    # https://confit.readthedocs.io/en/latest/#search-paths
+    # https://github.com/beetbox/confuse/blob/v1.4.0/confuse/core.py#L555
+    config = confuse.Configuration("pyvast_threatbus")
     config.set_args(args)
     if args.config:
         config.set_file(args.config)
@@ -642,9 +659,13 @@ def main():
     try:
         validate_config(config)
     except Exception as e:
-        raise ValueError(f"Invalid config: {e}")
+        sys.exit(ValueError(f"Invalid config: {e}"))
+
+    if config["logging"].get(dict):
+        setup_logging_with_config(config["logging"])
+    else:
+        setup_logging_with_level(config["loglevel"].get(str))
 
-    setup_logging(config["loglevel"].get(str))
     while True:
         try:
             asyncio.run(
@@ -657,6 +678,8 @@ def main():
                     config["retro_match_max_events"].get(),
                     config["unflatten"].get(),
                     config["max_background_tasks"].get(),
+                    config["metrics"]["interval"].get(),
+                    config["metrics"]["filename"].get(),
                     config["transform_context"].get(),
                     config["sink"].get(),
                 )
